// EERPOM Functions for Mouse Jiggler

#include <EEPROM.h>

// New Defines for mj_eeprom.h
#define MJ_EEPROM_SIZE 1024             // We assume a adruino Leonardo with Atmega32u4 - 1024 Bytes EEPROM

#define MJ_EEPROM_CRC 0                 // unsigned long CRC over EEPROM values
#define MJ_EEPROM_AUTOTYPE_DELAY 4      // unsigned long Delay after PowerOn before Autotyping
#define MJ_EEPROM_AUTOTYPE_1 8          // byte Autotype Action after PowerOn (0 = off, 1-14 = # of text in eeprom to use)
#define MJ_EEPROM_AUTOTYPE_2 9          // byte Autotype Action after PowerOn (0 = off, 1-14 = # of text in eeprom to use)
#define MJ_EEPROM_OSMODE 10             // unsigned char Operating System Mode - 0 = Windows & Linux Text Console. 1 = Linux GTK+ aso.
#define MJ_EEPROM_JIGGLER_ENABLE 11     // bool Enable MouseJiggler on Boot
#define MJ_EEPROM_JIGGLER_PIXEL 12      // byte Pixel Range for Jiggler
#define MJ_EEPROM_JIGGLER_MIN 13        // unsigned int Minimum Wait Time between Jiggles
#define MJ_EEPROM_JIGGLER_MAX 15        // unsigned int Maximum Wair Time between Jiggles

#define MJ_EEPROM_LEDMODE 17            // byte LED Blinkmode
#define MJ_EEPROM_LEDOPTION 18          // unsigned int LED Blink Options
#define MJ_EEPROM_LEDTIMING_1 20        // unsigned int 1st LED Timing Modifier
#define MJ_EEPROM_LEDTIMING_2 22        // unsigned int 2nd LED Timing Modifier
#define MJ_EEPROM_LEDTIMING_3 24        // unsigned int 3rd LED Timing Modifier
#define MJ_EEPROM_LEDTIMING_4 26        // unsigned int 4th LED Timing Modifier
#define MJ_EEPROM_LEDPIN_1 28           // byte LED Pin1
#define MJ_EEPROM_LEDPIN_2 29           // byte LED Pin2
#define MJ_EEPROM_RESERVED_1 30         // 1 byte Reserved
#define MJ_EEPROM_RESERVED_2 31         // 1 byte Reserved

#define MJ_EEPROM_DEFAULT_KEYDELAY 32   // unsigned char index into mj_keydelay[0-9];
#define MJ_EEPROM_RESERVED_3 33         // 1 byte Reserved
#define MJ_EEPROM_RESERVED_4 34         // 1 byte Reserved
#define MJ_EEPROM_RESERVED_5 35         // 1 byte Reserved
#define MJ_EEPROM_KEYDELAY_0 36			    // unsigned long ms between key presses Value '0'
#define MJ_EEPROM_KEYDELAY_1 40			    // unsigned long ms between key presses Value '1'
#define MJ_EEPROM_KEYDELAY_2 44			    // unsigned long ms between key presses Value '2'
#define MJ_EEPROM_KEYDELAY_3 48			    // unsigned long ms between key presses Value '3'
#define MJ_EEPROM_KEYDELAY_4 52			    // unsigned long ms between key presses Value '4'
#define MJ_EEPROM_KEYDELAY_5 56			    // unsigned long ms between key presses Value '5'
#define MJ_EEPROM_KEYDELAY_6 60			    // unsigned long ms between key presses Value '6'
#define MJ_EEPROM_KEYDELAY_7 64			    // unsigned long ms between key presses Value '7'
#define MJ_EEPROM_KEYDELAY_8 68			    // unsigned long ms between key presses Value '8'
#define MJ_EEPROM_KEYDELAY_9 72			    // unsigned long ms between key presses Value '9'

#define MJ_EEPROM_KEYPIN_1 80           // byte KEY1 Pin
#define MJ_EEPROM_KEYPIN_2 81           // byte KEY2 Pin
#define MJ_EEPROM_KEYPIN_3 82           // byte KEY3 Pin
#define MJ_EEPROM_KEYPIN_4 83           // byte KEY4 Pin
#define MJ_EEPROM_KEYPIN_5 84           // byte KEY5 Pin
#define MJ_EEPROM_KEYPIN_6 85           // byte KEY6 Pin
#define MJ_EEPROM_KEYPIN_7 86           // byte KEY7 Pin
#define MJ_EEPROM_KEYPIN_8 87           // byte KEY8 Pin
#define MJ_EEPROM_KEYPIN_9 88           // byte KEY9 Pin
#define MJ_EEPROM_KEY_1_PRESS_ACTION_1  89 // byte Key 1 press Action 1
#define MJ_EEPROM_KEY_1_PRESS_ACTION_2  90 // byte Key 1 press Action 2
#define MJ_EEPROM_KEY_1_RELEASE_ACTION_1   91 // byte Key 1 release Action 1
#define MJ_EEPROM_KEY_1_RELEASE_ACTION_2   92 // byte Key 1 release Action 2
#define MJ_EEPROM_KEY_2_PRESS_ACTION_1  93 // ...
#define MJ_EEPROM_KEY_2_PRESS_ACTION_2  94
#define MJ_EEPROM_KEY_2_RELEASE_ACTION_1   95
#define MJ_EEPROM_KEY_2_RELEASE_ACTION_2   96
#define MJ_EEPROM_KEY_3_PRESS_ACTION_1  97
#define MJ_EEPROM_KEY_3_PRESS_ACTION_2  98
#define MJ_EEPROM_KEY_3_RELEASE_ACTION_1   99
#define MJ_EEPROM_KEY_3_RELEASE_ACTION_2   100
#define MJ_EEPROM_KEY_4_PRESS_ACTION_1  101
#define MJ_EEPROM_KEY_4_PRESS_ACTION_2  102
#define MJ_EEPROM_KEY_4_RELEASE_ACTION_1   103
#define MJ_EEPROM_KEY_4_RELEASE_ACTION_2   104
#define MJ_EEPROM_KEY_5_PRESS_ACTION_1  105
#define MJ_EEPROM_KEY_5_PRESS_ACTION_2  106
#define MJ_EEPROM_KEY_5_RELEASE_ACTION_1   107
#define MJ_EEPROM_KEY_5_RELEASE_ACTION_2   108
#define MJ_EEPROM_KEY_6_PRESS_ACTION_1  109
#define MJ_EEPROM_KEY_6_PRESS_ACTION_2  110
#define MJ_EEPROM_KEY_6_RELEASE_ACTION_1   111
#define MJ_EEPROM_KEY_6_RELEASE_ACTION_2   112
#define MJ_EEPROM_KEY_7_PRESS_ACTION_1  113
#define MJ_EEPROM_KEY_7_PRESS_ACTION_2  114
#define MJ_EEPROM_KEY_7_RELEASE_ACTION_1   115
#define MJ_EEPROM_KEY_7_RELEASE_ACTION_2   116
#define MJ_EEPROM_KEY_8_PRESS_ACTION_1  117
#define MJ_EEPROM_KEY_8_PRESS_ACTION_2  118
#define MJ_EEPROM_KEY_8_RELEASE_ACTION_1   119
#define MJ_EEPROM_KEY_8_RELEASE_ACTION_2   120
#define MJ_EEPROM_KEY_9_PRESS_ACTION_1  121
#define MJ_EEPROM_KEY_9_PRESS_ACTION_2  122
#define MJ_EEPROM_KEY_9_RELEASE_ACTION_1   123
#define MJ_EEPROM_KEY_9_RELEASE_ACTION_2   124

#define MJ_EEPROM_RESERVED_6 125         // Reserved
#define MJ_EEPROM_RESERVED_7 126         // Reserved
#define MJ_EEPROM_RESERVED_8 127         // Reserved
//...
#define MJ_EEPROM_CONTENT_1_USEDBYTES 128  // byte
#define MJ_EEPROM_CONTENT_1_START     129
#define MJ_EEPROM_CONTENT_1_END       191
#define MJ_EEPROM_CONTENT_1_SIZE (MJ_EEPROM_CONTENT_1_END - MJ_EEPROM_CONTENT_1_START)
#define MJ_EEPROM_CONTENT_2_USEDBYTES 192
#define MJ_EEPROM_CONTENT_2_START     193
#define MJ_EEPROM_CONTENT_2_END       255
#define MJ_EEPROM_CONTENT_2_SIZE (MJ_EEPROM_CONTENT_2_END - MJ_EEPROM_CONTENT_2_START)
#define MJ_EEPROM_CONTENT_3_USEDBYTES 256
#define MJ_EEPROM_CONTENT_3_START     257
#define MJ_EEPROM_CONTENT_3_END       319
#define MJ_EEPROM_CONTENT_3_SIZE (MJ_EEPROM_CONTENT_3_END - MJ_EEPROM_CONTENT_3_START)
#define MJ_EEPROM_CONTENT_4_USEDBYTES 320
#define MJ_EEPROM_CONTENT_4_START     321
#define MJ_EEPROM_CONTENT_4_END       383
#define MJ_EEPROM_CONTENT_4_SIZE (MJ_EEPROM_CONTENT_4_END - MJ_EEPROM_CONTENT_4_START)
#define MJ_EEPROM_CONTENT_5_USEDBYTES 384
#define MJ_EEPROM_CONTENT_5_START     385
#define MJ_EEPROM_CONTENT_5_END       447
#define MJ_EEPROM_CONTENT_5_SIZE (MJ_EEPROM_CONTENT_5_END - MJ_EEPROM_CONTENT_5_START)
#define MJ_EEPROM_CONTENT_6_USEDBYTES 448
#define MJ_EEPROM_CONTENT_6_START     449
#define MJ_EEPROM_CONTENT_6_END       511
#define MJ_EEPROM_CONTENT_6_SIZE (MJ_EEPROM_CONTENT_6_END - MJ_EEPROM_CONTENT_6_START)
#define MJ_EEPROM_CONTENT_7_USEDBYTES 512
#define MJ_EEPROM_CONTENT_7_START     513
#define MJ_EEPROM_CONTENT_7_END       575
#define MJ_EEPROM_CONTENT_7_SIZE (MJ_EEPROM_CONTENT_7_END - MJ_EEPROM_CONTENT_7_START)
#define MJ_EEPROM_CONTENT_8_USEDBYTES 576
#define MJ_EEPROM_CONTENT_8_START     577
#define MJ_EEPROM_CONTENT_8_END       639
#define MJ_EEPROM_CONTENT_8_SIZE (MJ_EEPROM_CONTENT_8_END - MJ_EEPROM_CONTENT_8_START)
#define MJ_EEPROM_CONTENT_9_USEDBYTES 640
#define MJ_EEPROM_CONTENT_9_START     641
#define MJ_EEPROM_CONTENT_9_END       703
#define MJ_EEPROM_CONTENT_9_SIZE (MJ_EEPROM_CONTENT_9_END - MJ_EEPROM_CONTENT_9_START)
#define MJ_EEPROM_CONTENT_10_USEDBYTES 704
#define MJ_EEPROM_CONTENT_10_START     705
#define MJ_EEPROM_CONTENT_10_END       767
#define MJ_EEPROM_CONTENT_10_SIZE (MJ_EEPROM_CONTENT_10_END - MJ_EEPROM_CONTENT_10_START)
#define MJ_EEPROM_CONTENT_11_USEDBYTES 768
#define MJ_EEPROM_CONTENT_11_START     769
#define MJ_EEPROM_CONTENT_11_END       831
#define MJ_EEPROM_CONTENT_11_SIZE (MJ_EEPROM_CONTENT_11_END - MJ_EEPROM_CONTENT_11_START)
#define MJ_EEPROM_CONTENT_12_USEDBYTES 832
#define MJ_EEPROM_CONTENT_12_START     833
#define MJ_EEPROM_CONTENT_12_END       895
#define MJ_EEPROM_CONTENT_12_SIZE (MJ_EEPROM_CONTENT_12_END - MJ_EEPROM_CONTENT_12_START)
#define MJ_EEPROM_CONTENT_13_USEDBYTES 896
#define MJ_EEPROM_CONTENT_13_START     897
#define MJ_EEPROM_CONTENT_13_END       959
#define MJ_EEPROM_CONTENT_13_SIZE (MJ_EEPROM_CONTENT_13_END - MJ_EEPROM_CONTENT_13_START)
#define MJ_EEPROM_CONTENT_14_USEDBYTES 960
#define MJ_EEPROM_CONTENT_14_START     961
#define MJ_EEPROM_CONTENT_14_END       1023
#define MJ_EEPROM_CONTENT_14_SIZE (MJ_EEPROM_CONTENT_14_END - MJ_EEPROM_CONTENT_14_START)

#define MJ_EEPROM_CRCAREA_START 4
#define MJ_EEPROM_CRCAREA_END 1023

#define MJ_DEFAULT_AUTOTYPE_1 0
#define MJ_DEFAULT_AUTOTYPE_2 0
#define MJ_DEFAULT_AUTOTYPE_DELAY 0
#define MJ_DEFAULT_OSMODE 0
#define MJ_DEFAULT_JIGGLER_ENABLE 1
#define MJ_DEFAULT_JIGGLER_PIXEL 3
#define MJ_DEFAULT_JIGGLER_MIN 5000
#define MJ_DEFAULT_JIGGLER_MAX 15000
#define MJ_DEFAULT_LEDMODE 4
#define MJ_DEFAULT_LEDOPTION 0
#define MJ_DEFAULT_LEDTIMING_1 10000
#define MJ_DEFAULT_LEDTIMING_2 10000
#define MJ_DEFAULT_LEDTIMING_3 10000
#define MJ_DEFAULT_LEDTIMING_4 10000
#define MJ_DEFAULT_LEDPIN_1 LED_BUILTIN
#define MJ_DEFAULT_LEDPIN_2 LED_BUILTIN_RX
#define MJ_DEFAULT_KEYDELAY_DEFAULT 2
#define MJ_DEFAULT_KEYDELAY_0 2
#define MJ_DEFAULT_KEYDELAY_1 50
#define MJ_DEFAULT_KEYDELAY_2 100
#define MJ_DEFAULT_KEYDELAY_3 250
#define MJ_DEFAULT_KEYDELAY_4 400
#define MJ_DEFAULT_KEYDELAY_5 500
#define MJ_DEFAULT_KEYDELAY_6 750
#define MJ_DEFAULT_KEYDELAY_7 1000
#define MJ_DEFAULT_KEYDELAY_8 2000
#define MJ_DEFAULT_KEYDELAY_9 5000
#define MJ_DEFAULT_KEYPIN_1 9
#define MJ_DEFAULT_KEYPIN_2 11
#define MJ_DEFAULT_KEYPIN_3 10
#define MJ_DEFAULT_KEYPIN_4 8
#define MJ_DEFAULT_KEYPIN_5 0xff
#define MJ_DEFAULT_KEYPIN_6 0xff
#define MJ_DEFAULT_KEYPIN_7 0xff
#define MJ_DEFAULT_KEYPIN_8 0xff
#define MJ_DEFAULT_KEYPIN_9 0xff
#define MJ_DEFAULT_KEY_1_PRESS_ACTION_1 252 // Toggle Jiggler
#define MJ_DEFAULT_KEY_1_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_1_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_1_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_2_PRESS_ACTION_1 32 // Shutdown Windows
#define MJ_DEFAULT_KEY_2_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_2_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_2_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_3_PRESS_ACTION_1 34 // Lock Windows Screen
#define MJ_DEFAULT_KEY_3_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_3_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_3_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_4_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_4_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_4_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_4_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_5_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_5_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_5_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_5_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_6_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_6_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_6_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_6_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_7_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_7_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_7_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_7_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_8_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_8_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_8_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_8_RELEASE_ACTION_2  0
#define MJ_DEFAULT_KEY_9_PRESS_ACTION_1 0
#define MJ_DEFAULT_KEY_9_PRESS_ACTION_2 0
#define MJ_DEFAULT_KEY_9_RELEASE_ACTION_1  0
#define MJ_DEFAULT_KEY_9_RELEASE_ACTION_2  0

// Global Variables which will be initialized by eeprom contents
byte      mj_autotype_1;
byte      mj_autotype_2;
unsigned long mj_autotype_delay;
unsigned char mj_osmode;
bool 		  mj_jiggler_enable;
byte       mj_jiggler_pixel;
unsigned int mj_jiggler_min;
unsigned int mj_jiggler_max;
unsigned int  mj_ledmode;
unsigned int  mj_ledoption;
unsigned int  mj_ledtiming[4];
bool mj_ledtiming_changed = true;
unsigned int mj_ledpin1;
unsigned int mj_ledpin2;
unsigned char mj_keydelay_default;
unsigned long mj_keydelay[10];
unsigned int  mj_usedbytes[14];
unsigned int mj_keypin[9];
byte mj_action[9][2][2];

byte mj_eeprom_read_byte(int index) {
  if (index < MJ_EEPROM_SIZE) {
    return (EEPROM.read(index));
  } else {
    return (0);
  }
}

unsigned int mj_eeprom_read_int(int index) {
  unsigned int value = 0;
  if (index < MJ_EEPROM_SIZE) {
    value = (unsigned int)(EEPROM.read(index) << 8) + (unsigned int)(EEPROM.read(index + 1));
  }
  return (value);
}

unsigned long mj_eeprom_read_long(int index) {
  if (index < (MJ_EEPROM_SIZE - 3)) {
    return ( ((unsigned long) mj_eeprom_read_int(index) << 16) + ((unsigned long) mj_eeprom_read_int(index + 2)) );
  }
  return (0);
}

// Read a value of type 1=byte,2=int or 4=long from eeprom. Maybe not needed function.
unsigned long mj_eeprom_read_value(int index, int type) {
  switch (type) {
    case 1:  // byte
      return ((unsigned long) mj_eeprom_read_byte(index));
      break;
    case 2:  // int
      return ((unsigned long) mj_eeprom_read_int(index));
      break;
    case 4:
    default:
      return ((unsigned long) mj_eeprom_read_long(index));
      break;
  }
}

int mj_eeprom_write_byte(int index, byte b) {
  if (index < MJ_EEPROM_SIZE) {
    EEPROM.update(index, b);
    return (true);
  } else {
    return (false);
  }
}

int mj_eeprom_write_int(int index, unsigned int i) {
  if (index < (MJ_EEPROM_SIZE - 1)) {
    EEPROM.update(index, (byte) (i >> 8));
    EEPROM.update(index + 1, (byte) i & 0x00ff);
    return (true);
  } else {
    return (false);
  }
}

int mj_eeprom_write_long(int index, unsigned long l) {
  if (index < (MJ_EEPROM_SIZE - 3)) {
    EEPROM.update(index, (byte)(l >> 24));
    EEPROM.update(index + 1, (byte)((l & (unsigned long) 0x00ff0000) >> 16));
    EEPROM.update(index + 2, (byte)((l & (unsigned long) 0x0000ff00) >> 8));
    EEPROM.update(index + 3, (byte)(l & (unsigned long)  0x000000ff));
    return (true);
  } else {
    return (false);
  }
}

// Update a value of type 1=byte,2=int or 4=long in eeprom.
int mj_eeprom_write_value(int index, int type, unsigned long value) {
  switch (type) {
    case 1:  // byte
      return (mj_eeprom_write_byte(index, (unsigned char) value));
      break;
    case 2:  // int
      return (mj_eeprom_write_int(index, (unsigned int) value));
      break;
    case 4:  // long
    default:
      return (mj_eeprom_write_long(index, (unsigned long) value));
      break;
  }
}

// Calculate a CRC over the given Range in EEPROM
unsigned long mj_eeprom_calc_crc(int index_start, int index_end) {
  const unsigned long crc_table[16] = {
    0x00000000, 0x1db71064, 0x3b6e20c8, 0x26d930ac,
    0x76dc4190, 0x6b6b51f4, 0x4db26158, 0x5005713c,
    0xedb88320, 0xf00f9344, 0xd6d6a3e8, 0xcb61b38c,
    0x9b64c2b0, 0x86d3d2d4, 0xa00ae278, 0xbdbdf21c
  };
  unsigned long crc = ~0L;
  int i;
  if (index_start < MJ_EEPROM_SIZE && index_end < MJ_EEPROM_SIZE && index_start <= index_end) {
    for (i = index_start ; i < index_end ; ++i) {
      crc = crc_table[(crc ^ EEPROM[i]) & 0x0f] ^ (crc >> 4);
      crc = crc_table[(crc ^ (EEPROM[i] >> 4)) & 0x0f] ^ (crc >> 4);
      crc = ~crc;
    }
  }
  return (crc);
}

// Update a value of type 1=byte,2=int or 4=long in eeprom and updates CRC, too.
int mj_eeprom_update_value(int index, int type, unsigned long value) {
  unsigned long crc;
  mj_eeprom_write_value(index, type, value);
  crc = mj_eeprom_calc_crc(MJ_EEPROM_CRCAREA_START, MJ_EEPROM_CRCAREA_END);
  mj_eeprom_write_long(MJ_EEPROM_CRC, crc);
  return (1);
}

// Wipes all contents of eeprom
void mj_eeprom_wipe() {
  int i;
  for (i = 0; i < MJ_EEPROM_SIZE; i++) {
    EEPROM.update(i, 0xff);
  }
}

// Dumps eeprom Contents to Serial Console
void mj_eeprom_dump_serial() {
  int i, j, b, len;
  len = EEPROM.length();
  Serial.print(F("0x000  "));
  for (i = 0; i < len; i++) {
    b = EEPROM[i];
    if (b < 16) {
      Serial.write('0');
    }
    Serial.print(b, HEX);
    Serial.write(' ');
    if ((i + 1) % 4 == 0 ) {
      Serial.print(F(" "));
    }
    if ((i + 1) % 16 == 0) {
      Serial.write(' ');
      for (j = (i - 15); j <= i; j++) {
        b = EEPROM[j];
        if (b >= 32 && b <= 127) {
          Serial.write(b);
        } else {
          Serial.write('.');
        }
      }
      if ((i + 1) < len) {
        Serial.println();
        Serial.print(F("0x"));
        if (i < 255) {
          Serial.write('0');
        }
        Serial.print((i + 1) / 16, HEX);
        Serial.print(F("0  "));
      } else {
        Serial.println();
      }
    }
  }
}

// Initialize EEPROM Defaults
void mj_eeprom_init_defaults() {
  // Just write the values bulk into eeprom
  mj_eeprom_write_byte(MJ_EEPROM_AUTOTYPE_1, MJ_DEFAULT_AUTOTYPE_1);
  mj_eeprom_write_byte(MJ_EEPROM_AUTOTYPE_2, MJ_DEFAULT_AUTOTYPE_2);
  mj_eeprom_write_long(MJ_EEPROM_AUTOTYPE_DELAY, MJ_DEFAULT_AUTOTYPE_DELAY);
  mj_eeprom_write_byte(MJ_EEPROM_OSMODE, MJ_DEFAULT_OSMODE);
  mj_eeprom_write_byte(MJ_EEPROM_JIGGLER_ENABLE, MJ_DEFAULT_JIGGLER_ENABLE);
  mj_eeprom_write_byte(MJ_EEPROM_JIGGLER_PIXEL, MJ_DEFAULT_JIGGLER_PIXEL);
  mj_eeprom_write_int(MJ_EEPROM_JIGGLER_MIN, MJ_DEFAULT_JIGGLER_MIN);
  mj_eeprom_write_int(MJ_EEPROM_JIGGLER_MAX, MJ_DEFAULT_JIGGLER_MAX);
  mj_eeprom_write_int(MJ_EEPROM_LEDMODE, MJ_DEFAULT_LEDMODE);
  mj_eeprom_write_int(MJ_EEPROM_LEDOPTION, MJ_DEFAULT_LEDOPTION);
  mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_1, MJ_DEFAULT_LEDTIMING_1);
  mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_2, MJ_DEFAULT_LEDTIMING_2);
  mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_3, MJ_DEFAULT_LEDTIMING_3);
  mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_4, MJ_DEFAULT_LEDTIMING_4);
  mj_eeprom_write_byte(MJ_EEPROM_LEDPIN_1, MJ_DEFAULT_LEDPIN_1);
  mj_eeprom_write_byte(MJ_EEPROM_LEDPIN_2, MJ_DEFAULT_LEDPIN_2);
  mj_eeprom_write_byte(MJ_EEPROM_DEFAULT_KEYDELAY, MJ_DEFAULT_KEYDELAY_DEFAULT);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_0, MJ_DEFAULT_KEYDELAY_0);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_1, MJ_DEFAULT_KEYDELAY_1);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_2, MJ_DEFAULT_KEYDELAY_2);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_3, MJ_DEFAULT_KEYDELAY_3);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_4, MJ_DEFAULT_KEYDELAY_4);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_5, MJ_DEFAULT_KEYDELAY_5);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_6, MJ_DEFAULT_KEYDELAY_6);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_7, MJ_DEFAULT_KEYDELAY_7);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_8, MJ_DEFAULT_KEYDELAY_8);
  mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_9, MJ_DEFAULT_KEYDELAY_9);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_1, MJ_DEFAULT_KEYPIN_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_2, MJ_DEFAULT_KEYPIN_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_3, MJ_DEFAULT_KEYPIN_3);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_4, MJ_DEFAULT_KEYPIN_4);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_5, MJ_DEFAULT_KEYPIN_5);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_6, MJ_DEFAULT_KEYPIN_6);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_7, MJ_DEFAULT_KEYPIN_7);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_8, MJ_DEFAULT_KEYPIN_8);
  mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_9, MJ_DEFAULT_KEYPIN_9);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_1_PRESS_ACTION_1, MJ_DEFAULT_KEY_1_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_1_PRESS_ACTION_2, MJ_DEFAULT_KEY_1_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_1_RELEASE_ACTION_1, MJ_DEFAULT_KEY_1_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_1_RELEASE_ACTION_2, MJ_DEFAULT_KEY_1_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_2_PRESS_ACTION_1, MJ_DEFAULT_KEY_2_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_2_PRESS_ACTION_2, MJ_DEFAULT_KEY_2_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_2_RELEASE_ACTION_1, MJ_DEFAULT_KEY_2_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_2_RELEASE_ACTION_2, MJ_DEFAULT_KEY_2_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_3_PRESS_ACTION_1, MJ_DEFAULT_KEY_3_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_3_PRESS_ACTION_2, MJ_DEFAULT_KEY_3_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_3_RELEASE_ACTION_1, MJ_DEFAULT_KEY_3_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_3_RELEASE_ACTION_2, MJ_DEFAULT_KEY_3_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_4_PRESS_ACTION_1, MJ_DEFAULT_KEY_4_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_4_PRESS_ACTION_2, MJ_DEFAULT_KEY_4_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_4_RELEASE_ACTION_1, MJ_DEFAULT_KEY_4_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_4_RELEASE_ACTION_2, MJ_DEFAULT_KEY_4_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_5_PRESS_ACTION_1, MJ_DEFAULT_KEY_5_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_5_PRESS_ACTION_2, MJ_DEFAULT_KEY_5_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_5_RELEASE_ACTION_1, MJ_DEFAULT_KEY_5_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_5_RELEASE_ACTION_2, MJ_DEFAULT_KEY_5_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_6_PRESS_ACTION_1, MJ_DEFAULT_KEY_6_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_6_PRESS_ACTION_2, MJ_DEFAULT_KEY_6_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_6_RELEASE_ACTION_1, MJ_DEFAULT_KEY_6_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_6_RELEASE_ACTION_2, MJ_DEFAULT_KEY_6_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_7_PRESS_ACTION_1, MJ_DEFAULT_KEY_7_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_7_PRESS_ACTION_2, MJ_DEFAULT_KEY_7_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_7_RELEASE_ACTION_1, MJ_DEFAULT_KEY_7_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_7_RELEASE_ACTION_2, MJ_DEFAULT_KEY_7_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_8_PRESS_ACTION_1, MJ_DEFAULT_KEY_8_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_8_PRESS_ACTION_2, MJ_DEFAULT_KEY_8_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_8_RELEASE_ACTION_1, MJ_DEFAULT_KEY_8_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_8_RELEASE_ACTION_2, MJ_DEFAULT_KEY_8_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_9_PRESS_ACTION_1, MJ_DEFAULT_KEY_9_PRESS_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_9_PRESS_ACTION_2, MJ_DEFAULT_KEY_9_PRESS_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_9_RELEASE_ACTION_1, MJ_DEFAULT_KEY_9_RELEASE_ACTION_1);
  mj_eeprom_write_byte(MJ_EEPROM_KEY_9_RELEASE_ACTION_2, MJ_DEFAULT_KEY_9_RELEASE_ACTION_2);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_1_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_1_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_2_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_2_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_3_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_3_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_4_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_4_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_5_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_5_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_6_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_6_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_7_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_7_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_8_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_8_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_9_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_9_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_10_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_10_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_11_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_11_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_12_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_12_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_13_USEDBYTES, 0);
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_13_START, '\0');
  mj_eeprom_write_byte(MJ_EEPROM_CONTENT_14_USEDBYTES, 0);
  mj_eeprom_update_value(MJ_EEPROM_CONTENT_14_START, 1, '\0');  // Last update re-calculates crc, too
}

// Check EEPROM and init if CRC is incorrect. After that reads Values into global variables
int mj_eeprom_check_crc() {
  if (mj_eeprom_calc_crc(MJ_EEPROM_CRCAREA_START, MJ_EEPROM_CRCAREA_END) != mj_eeprom_read_long(MJ_EEPROM_CRC)) {
    return (false);
  }
  else {
    return (true);
  }
}

void mj_eeprom_get_config() {
  int a, b, c;
  // Read info from EERPOM into global Variables
  mj_autotype_1 = mj_eeprom_read_byte(MJ_EEPROM_AUTOTYPE_1);
  mj_autotype_2 = mj_eeprom_read_byte(MJ_EEPROM_AUTOTYPE_2);
  mj_autotype_delay = mj_eeprom_read_long(MJ_EEPROM_AUTOTYPE_DELAY);
  mj_osmode = mj_eeprom_read_byte(MJ_EEPROM_OSMODE);
  mj_jiggler_enable = mj_eeprom_read_byte(MJ_EEPROM_JIGGLER_ENABLE) & 0x01;
  mj_jiggler_pixel = (byte) mj_eeprom_read_byte(MJ_EEPROM_JIGGLER_PIXEL);
  mj_jiggler_min = mj_eeprom_read_int(MJ_EEPROM_JIGGLER_MIN);
  mj_jiggler_max = mj_eeprom_read_int(MJ_EEPROM_JIGGLER_MAX);
  mj_ledmode = mj_eeprom_read_int(MJ_EEPROM_LEDMODE);
  mj_ledoption = mj_eeprom_read_int(MJ_EEPROM_LEDOPTION);
  for (a = 0; a < 4; a++) {
    mj_ledtiming[a] = mj_eeprom_read_int(MJ_EEPROM_LEDTIMING_1 + (a * 2));
  }
  mj_ledtiming_changed = true;
  mj_ledpin1 = (int) mj_eeprom_read_byte(MJ_EEPROM_LEDPIN_1);
  mj_ledpin2 = (int) mj_eeprom_read_byte(MJ_EEPROM_LEDPIN_2);
  mj_keydelay_default = mj_eeprom_read_byte(MJ_EEPROM_DEFAULT_KEYDELAY);
  for (a = 0; a < 10; a++) {
    mj_keydelay[a] = mj_eeprom_read_long(MJ_EEPROM_KEYDELAY_0 + (a * 4));
  }
  mj_usedbytes[0] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_1_USEDBYTES);
  mj_usedbytes[1] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_2_USEDBYTES);
  mj_usedbytes[2] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_3_USEDBYTES);
  mj_usedbytes[3] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_4_USEDBYTES);
  mj_usedbytes[4] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_5_USEDBYTES);
  mj_usedbytes[5] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_6_USEDBYTES);
  mj_usedbytes[6] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_7_USEDBYTES);
  mj_usedbytes[7] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_8_USEDBYTES);
  mj_usedbytes[8] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_9_USEDBYTES);
  mj_usedbytes[9] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_10_USEDBYTES);
  mj_usedbytes[10] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_11_USEDBYTES);
  mj_usedbytes[11] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_12_USEDBYTES);
  mj_usedbytes[12] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_13_USEDBYTES);
  mj_usedbytes[13] = (int) mj_eeprom_read_byte(MJ_EEPROM_CONTENT_14_USEDBYTES);
  for (a = 0; a < 10; a++) {
    mj_keypin[a] = (int) mj_eeprom_read_byte(MJ_EEPROM_KEYPIN_1 + a);
  }
  for (a = 0; a < 9; a++) {
    for (b = 0; b < 2; b++) {
      for (c = 0; c < 2; c++) {
        mj_action[a][b][c] = mj_eeprom_read_byte(MJ_EEPROM_KEY_1_PRESS_ACTION_1 + ((a * 4) + (b * 2) + c));
      }
    }
  }
  if (mj_ledpin1 != 0xff) {
    pinMode(mj_ledpin1, OUTPUT);
  }
  if (mj_ledpin2 != 0xff) {
    pinMode(mj_ledpin2, OUTPUT);
  }
  for (a = 0; a < 10; a++) {
    if (mj_keypin[a] != 0xff ) {
      pinMode(mj_keypin[a], INPUT_PULLUP);
    }
  }
}

// Now more MJ Specific Code - above is mostly generic.

char mj_getvalue(int index, int type, char *c) {
  switch (type) {
    case 0:
      return (c[index]);
      break;
    case 0xff:
      return (pgm_read_byte(c + index));
      break;
    case 1:
      if (index < MJ_EEPROM_CONTENT_1_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_1_START + index));
      } else {
        return (0);
      }
      break;
    case 2:
      if (index < MJ_EEPROM_CONTENT_2_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_2_START + index));
      } else {
        return (0);
      }
      break;
    case 3:
      if (index < MJ_EEPROM_CONTENT_3_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_3_START + index));
      } else {
        return (0);
      }
      break;
    case 4:
      if (index < MJ_EEPROM_CONTENT_4_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_4_START + index));
      } else {
        return (0);
      }
      break;
    case 5:
      if (index < MJ_EEPROM_CONTENT_5_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_5_START + index));
      } else {
        return (0);
      }
      break;
    case 6:
      if (index < MJ_EEPROM_CONTENT_6_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_6_START + index));
      } else {
        return (0);
      }
      break;
    case 7:
      if (index < MJ_EEPROM_CONTENT_7_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_7_START + index));
      } else {
        return (0);
      }
      break;
    case 8:
      if (index < MJ_EEPROM_CONTENT_8_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_8_START + index));
      } else {
        return (0);
      }
      break;
    case 9:
      if (index < MJ_EEPROM_CONTENT_9_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_9_START + index));
      } else {
        return (0);
      }
      break;
    case 10:
      if (index < MJ_EEPROM_CONTENT_10_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_10_START + index));
      } else {
        return (0);
      }
      break;
    case 11:
      if (index < MJ_EEPROM_CONTENT_11_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_11_START + index));
      } else {
        return (0);
      }
      break;
    case 12:
      if (index < MJ_EEPROM_CONTENT_12_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_12_START + index));
      } else {
        return (0);
      }
      break;
    case 13:
      if (index < MJ_EEPROM_CONTENT_13_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_13_START + index));
      } else {
        return (0);
      }
      break;
    case 14:
    default:
      if (index < MJ_EEPROM_CONTENT_14_SIZE) {
        return (EEPROM.read(MJ_EEPROM_CONTENT_14_START + index));
      } else {
        return (0);
      }
      break;
  }
}

void put_eeprom( unsigned char *c) {
  int len, i = 0;

  len = strlen((const char*) c);
  // We never can print more than the Fifosize or Buffersize so it's useless to write more to EEPROM
  if (len >= MJ_FIFOSIZE) {
    len = MJ_FIFOSIZE - 1;
  }
  // Maybe we have more bytes than defined space in EEPROM - then limit to this, too
  if (len >= (MJ_EEPROM_CONTENT_1_END - MJ_EEPROM_CONTENT_1_START)) {
    len = MJ_EEPROM_CONTENT_1_END - MJ_EEPROM_CONTENT_1_START;
  }
  for (i = 0; i < len; i++) {
    EEPROM.update(MJ_EEPROM_CONTENT_1_START + i, c[i]);
  }
  EEPROM.update(MJ_EEPROM_CONTENT_1_START + i, 0);
  mj_eeprom_update_value(MJ_EEPROM_CONTENT_1_USEDBYTES, 2, (unsigned long) len);
}
