// LED Options Parser
// p = Pin Assignment. Example atlp113 sets LED1 Pin to 13. Pin number may not be higher than available pins
//           New Pin will be checked for collision with other LED Pin and Button Pin.
//     atlps Shows Pin Assingment
//     atlpS Saves Pin Assignment to EEPROM
//     atlpr resets Pins to default if no Collision with key pins
// t = LED Timing-Modifier. atlt110000 sets the Modifier 1 to 10000 (default), atlts Shows Timings, atltS Saves Timings to EEPROM, atltr restores Default Timings
// m = LED Mode. atlm<decimal value> sets the Mode, atlms shows mode, atlmS Saves Mode to EEPROM, atlmr restores Default LED Mode
// o = LED Option. atlo<4 hex digits> sets the options. atlor restores Default Options
void parseled(char *c, unsigned int num) {
  unsigned int value;
  int n;
  bool x;
  switch (c[0]) {
    case 'p':
      if (num >= 3) {
        switch (c[1]) {
          case '1':
          case '2':
            value = (unsigned int) mj_convert_dec_long(c + 2);
            Serial.print(value);
            Serial.println();
            // Check collision with other Pins already used (Key Inputs)
            x = true;
            for (n = 0; n < 4; n++) {  // Avoid collision with our 4 Key-Pins
              if (value == mj_keypin[n]) {
                x = false;
              }
            }
#ifndef LED_BUILTIN_RX   // For example Arduino Uno - if we'd use the RX/TX Pins for LED here we kill the serial communication. Leonardo works with 0/1
            if (value == 0 || value == 1) {
              x = false;
            }
#endif
            if (c[1] == '1') {
              if (value == mj_ledpin2) {
                x = false;
              }
            } else {
              if (value == mj_ledpin1) {
                x = false;
              }
            }
            if (x == true && value < NUM_DIGITAL_PINS ) {
              pinMode(value, OUTPUT);
              if (c[1] == '1') {
                digitalWrite(mj_ledpin1, 0);
                mj_ledpin1 = value;
              } else {
                digitalWrite(mj_ledpin2, 0);
                mj_ledpin2 = value;
              }
            } else if ( value == 0xff ) {
              if (c[1] == '1') {
                mj_ledpin1 = 0xff;
              } else {
                mj_ledpin2 = 0xff;
              }
            }
            break;
        }
      } else if (num == 2) {
        switch (c[1]) {
          case 'r': // Restore LED Default Pins
            if ((MJ_DEFAULT_LEDPIN_1 != mj_keypin[0]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[1]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[2]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[3]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[4]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[5]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[6]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[7]) &&
                (MJ_DEFAULT_LEDPIN_1 != mj_keypin[8])) {
              mj_ledpin1 = MJ_DEFAULT_LEDPIN_1;
            }
            if ((MJ_DEFAULT_LEDPIN_2 != mj_keypin[0]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[1]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[2]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[3]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[4]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[5]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[6]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[7]) &&
                (MJ_DEFAULT_LEDPIN_2 != mj_keypin[8])) {
              mj_ledpin2 = MJ_DEFAULT_LEDPIN_2;
            }
            if (mj_ledpin1 != 0xff) {
              pinMode(mj_ledpin1, OUTPUT);
            }
            if (mj_ledpin2 != 0xff) {
              pinMode(mj_ledpin2, OUTPUT);
            }
            [[fallthrough]];
          case 's': // Show serial Pin Assignment
            Serial.print(F("PIN1:"));
            if (mj_ledpin1 != 0xff) {
              Serial.print(mj_ledpin1);
            } else {
              Serial.print(F("dis"));
            }
            Serial.print(F(" PIN2:"));
            if (mj_ledpin2 != 0xff) {
              Serial.println(mj_ledpin2);
            } else {
              Serial.println(F("dis"));
            }
            break;
          case 'S': // Save Serial Pin Assignment to EEPROM
            mj_eeprom_write_byte(MJ_EEPROM_LEDPIN_1, mj_ledpin1); // Save Pin 1
            mj_eeprom_update_value(MJ_EEPROM_LEDPIN_2, 1, mj_ledpin2); // Save Pin 2 plus recalculate + Update CRC
            break;
          default:
            break;
        }
      }
      break;
    case 't': // LED Timing Modifiers
      if (num >= 2) {
        switch (c[1]) {
          case '1':
          case '2':
          case '3':
          case '4':
            if (num > 2) {
              value = (unsigned int) mj_convert_dec_long(c + 2);
              n = c[1] - '1';
              mj_ledtiming[n] = value;
              mj_ledtiming_changed = true; // Force Recalculation of LED Timings
            }
            break;
          case 'r': // Restore Default Timings
            mj_ledtiming[0] = MJ_DEFAULT_LEDTIMING_1;
            mj_ledtiming[1] = MJ_DEFAULT_LEDTIMING_2;
            mj_ledtiming[2] = MJ_DEFAULT_LEDTIMING_3;
            mj_ledtiming[3] = MJ_DEFAULT_LEDTIMING_4;
            mj_ledtiming_changed = true;
            [[fallthrough]];
          case 's':
            Serial.print(F("LED Timing"));
            for (n = 0; n < 4; n++) {
              Serial.print(n + 1);
              Serial.print(F(":"));
              Serial.print(mj_ledtiming[n]);
              Serial.write(' ');
            }
            Serial.println();
            break;
          case 'S': // Save LED Timings to EEPROM
            mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_1, mj_ledtiming[0]);
            mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_2, mj_ledtiming[1]);
            mj_eeprom_write_int(MJ_EEPROM_LEDTIMING_3, mj_ledtiming[2]);
            mj_eeprom_update_value(MJ_EEPROM_LEDTIMING_4, 2, mj_ledtiming[3]);
            break;
          default:
            break;
        }
      }
      break;
    case 'm': // LED Mode
      if (num >= 2) {
        switch (c[1]) {
          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
            value = (unsigned int) mj_convert_dec_long(c + 1);
            mj_ledmode = value;
            mj_ledtiming_changed = true; // If LED Mode changes then recalculate Timings in LED Process
            break;
          case 'r':
            mj_ledmode = MJ_DEFAULT_LEDMODE;
            mj_ledtiming_changed = true; // If LED Mode changes then recalculate Timings in LED Process
            [[fallthrough]];
          case 's':
            Serial.print(F("LED Mode: "));
            Serial.println(mj_ledmode);
            break;
          case 'S':
            mj_eeprom_update_value(MJ_EEPROM_LEDMODE, 2, mj_ledmode);
            break;
          default:
            break;
        }
      }
      break;
    case 'o': // LED Options
      if (num >= 2) {
        switch (c[1]) {
          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
          case 'a':
          case 'b':
          case 'c':
          case 'd':
          case 'e':
          case 'f':
          case 'A':
          case 'B':
          case 'C':
          case 'D':
          case 'E':
          case 'F':
            value = (unsigned int) mj_convert_hex_long(c + 1);
            mj_ledoption = value;
            mj_ledtiming_changed = true; // If LED Option changes then recalculate Timings in LED Process
            break;
          case 'r':
            mj_ledoption = MJ_DEFAULT_LEDOPTION;
            mj_ledtiming_changed = true; // If LED Option changes then recalculate Timings in LED Process
            [[fallthrough]];
          case 's':
            Serial.print(F("LED Option: 0x"));
            Serial.println(mj_ledoption, HEX);
            break;
          case 'S':
            mj_eeprom_update_value(MJ_EEPROM_LEDOPTION, 2, mj_ledoption);
            break;
          default:
            break;
        }
      }
      break;
  }
}

// Commands ate...
// d = dump eeprom contents
// c = Calc CRC and show calculated and stored CRC
// C = Calc CRC and update stored CDC
// i,I,z = initialize EEPROM Contents - need to be given 3 times - i just rewrites known content, I wipes all, z only zeroes whole eeprom (0xff) but does not load defaults
// w,W = Write value b/c = byte, i = int, l = long to given address (3 hex digits) with value (1-8 digits). byte = 2, int = 4, long = 1-8 digits W with update CRC, w without update CRC
// r,R = Read value from address (same as w/W but without value) and show

void parseeeprom(char *c, unsigned int num) {
  unsigned long crc_calc, crc_stored;
  static int i = 2;
  unsigned long v;
  int t = 0, a = 0;
  switch (c[0]) {
    case 'i':
    case 'I':
    case 'z':
      ;
      break;
    default:
      i = 2;
      break;
  }
  switch (c[0]) {
    case 'd':
      mj_eeprom_dump_serial();
      break;
    case 'c':
      crc_calc = mj_eeprom_calc_crc(MJ_EEPROM_CRCAREA_START, MJ_EEPROM_CRCAREA_END);
      crc_stored = mj_eeprom_read_long(MJ_EEPROM_CRC);
      Serial.print(F("EEPROM CRC "));
      Serial.write((crc_calc == crc_stored) ? "valid" : "invalid");
      Serial.print(F(" stored: "));
      Serial.print(crc_stored);
      Serial.print(F(" calculated: "));
      Serial.println(crc_calc);
      break;
    case 'C':
      crc_calc = mj_eeprom_calc_crc(MJ_EEPROM_CRCAREA_START, MJ_EEPROM_CRCAREA_END);
      mj_eeprom_write_long(MJ_EEPROM_CRC, crc_calc);
      break;
    case 'i':
    case 'I':
    case 'z':
      if (i > 0) {
        Serial.print(F("Please repeat command "));
        Serial.print((i == 2) ? "twice" : "once");
        Serial.println(F(" to wipe EEPROM"));
        i--;
      } else {
        i = 2;
        if (c[0] == 'I' || c[0] == 'z') {
          Serial.println(F("Wiping"));
          mj_eeprom_wipe();
        }
        if (c[0] == 'I' || c[0] == 'i') {
          Serial.println(F("Initializing defaults"));
          mj_eeprom_init_defaults();
          mj_eeprom_get_config();
        }
      }
      break;
    case 'w':
    case 'W':  // Write wants to have 5,7 or 11 Hex digits depending on modifier: atwb 5, atwi 7, atwl 11. first 3 digits address, remaining value.
      if (num >= 6) {
        a = mj_convert_hex_int('0', c[2], c[3], c[4]);
      }
      switch (c[1]) {
        case 'b':
        case 'c': // One char
          if (num == 7) {
            v = (unsigned long) mj_convert_hex_char(c[5], c[6]);
            t = 1;
          }
          break;
        case 'i': // int
          if (num == 9) {
            v = (unsigned long) mj_convert_hex_int(c[5], c[6], c[7], c[8]);
            t = 2;
          }
          break;
        case 'l': // long
          if (num >= 6) {
            v = mj_convert_hex_long(c + 5);
            t = 4;
          }
          break;
      }
      if (t > 0) {
        switch (t) {
          case 1:
            mj_eeprom_write_byte(a, (byte) v);
            break;
          case 2:
            mj_eeprom_write_int(a, (unsigned int) v);
            break;
          case 4:
            mj_eeprom_write_long(a, v);
            break;
          default:
            break;
        }
        if (c[0] == 'W') {
          crc_calc = mj_eeprom_calc_crc(MJ_EEPROM_CRCAREA_START, MJ_EEPROM_CRCAREA_END);
          mj_eeprom_write_long(MJ_EEPROM_CRC, crc_calc);
        }
      }
      break;
    case 'r':
    case 'R':  // Read wants to have 3 hex digits address.
      if (num == 5) {
        a = mj_convert_hex_int('0', c[2], c[3], c[4]);
      }
      switch (c[1]) {
        case 'b':
        case 'c': // One char
          v = (unsigned long) mj_eeprom_read_byte(a);
          t = 1;
          break;
        case 'i': // int
          v = (unsigned long) mj_eeprom_read_int(a);
          t = 2;
          break;
        case 'l': // long
          v = mj_eeprom_read_long(a);
          t = 4;
          break;
        default:
          break;
      }
      if (t > 0) {
        Serial.print(F("Addr: "));
        Serial.print(a, HEX);
        Serial.print(F(" value: "));
        Serial.println(v, HEX);
      }
      break;
  }
}

// Command att...
// atts - print eeprom content to serial
// attk - write string after k to keyboard
// attw<1-e> - overwrite string '1'-'e' (1-14) in eeprom - if no string is given this will mark the string deleted
// attt<1-e> - type string in eeprom
// attd<0-9><decimal> - configure typedelay <0-9> to <decimal> ms
// attds - show active typedelays
// attdr - reset typedelays to default
// attdS - save active typedelays to EEPROM
// attar - reset autotype delay to default
// attas - show autotype delay
// attat<decimal> - set autotype delay to <decimal> in ms
// attad - disable autotyping
// atta1<num> - set autotype action #1
// atta2<num> - set autotype action #2
// attaS - Save autotype delay to EEPROM

void parsekeyboard(char *c, unsigned int num) {
  unsigned int i = 0, j, k, s;
  unsigned long value = 0;
  switch (c[0]) {
    case 'o':
      if (num == 2) {
        switch (c[1]) {
          case '0':
          case '1':
            mj_osmode = c[1] - '0';
            break;
          case 's':
            Serial.print(F("OSMode: "));
            Serial.println(mj_osmode);
            break;
          case 'r':
            mj_osmode = MJ_DEFAULT_OSMODE;
            break;
          case 'S':
            mj_eeprom_write_byte(MJ_EEPROM_OSMODE, mj_osmode);
            break;
        }
      }
      break;
    case 'a':
      if (num > 2 && c[1] == 't') {
        mj_autotype_delay = mj_convert_dec_long(c + 2);
      } else if (num == 2) {
        switch (c[1]) {
          case 'r':
            mj_autotype_delay = MJ_DEFAULT_AUTOTYPE_DELAY;
            mj_autotype_1 = MJ_DEFAULT_AUTOTYPE_1;
            mj_autotype_2 = MJ_DEFAULT_AUTOTYPE_2;
            [[fallthrough]];
          case 's':
            Serial.print(F("AutoType is "));
            Serial.print(mj_autotype_1 > 0 ? "en" : "dis");
            Serial.print(F("abled. Delay "));
            Serial.println(mj_autotype_delay);
            if (mj_autotype_1 > 0) {
              Serial.print(F("Action 1:"));
              Serial.print(mj_autotype_1);
              Serial.print(F(" Action 2:"));
              Serial.println(mj_autotype_2);
            }
            break;
          case 'S':
            mj_eeprom_write_byte(MJ_EEPROM_AUTOTYPE_1, mj_autotype_1);
            mj_eeprom_write_byte(MJ_EEPROM_AUTOTYPE_2, mj_autotype_2);
            mj_eeprom_update_value(MJ_EEPROM_AUTOTYPE_DELAY, 4, mj_autotype_delay);
            break;
          case 'd':
            mj_autotype_1 = 0;
            break;
          default:
            break;
        }
      } else if (num > 2 && (c[1] == '1' || c[1] == '2')) {
        value = mj_convert_dec_long(c + 2);
        if (c[1] == '1') {
          mj_autotype_1 = (byte) value;
        } else {
          mj_autotype_2 = (byte) value;
        }
      }
      break;
    case 's':
      for (j = 0; j < 14; j++) {
        switch (j + 1) {
          case 1: k = MJ_EEPROM_CONTENT_1_START; s = MJ_EEPROM_CONTENT_1_SIZE; break;
          case 2: k = MJ_EEPROM_CONTENT_2_START; s = MJ_EEPROM_CONTENT_2_SIZE; break;
          case 3: k = MJ_EEPROM_CONTENT_3_START; s = MJ_EEPROM_CONTENT_3_SIZE; break;
          case 4: k = MJ_EEPROM_CONTENT_4_START; s = MJ_EEPROM_CONTENT_4_SIZE; break;
          case 5: k = MJ_EEPROM_CONTENT_5_START; s = MJ_EEPROM_CONTENT_5_SIZE; break;
          case 6: k = MJ_EEPROM_CONTENT_6_START; s = MJ_EEPROM_CONTENT_6_SIZE; break;
          case 7: k = MJ_EEPROM_CONTENT_7_START; s = MJ_EEPROM_CONTENT_7_SIZE; break;
          case 8: k = MJ_EEPROM_CONTENT_8_START; s = MJ_EEPROM_CONTENT_8_SIZE; break;
          case 9: k = MJ_EEPROM_CONTENT_9_START; s = MJ_EEPROM_CONTENT_9_SIZE; break;
          case 10: k = MJ_EEPROM_CONTENT_10_START; s = MJ_EEPROM_CONTENT_10_SIZE; break;
          case 11: k = MJ_EEPROM_CONTENT_11_START; s = MJ_EEPROM_CONTENT_11_SIZE; break;
          case 12: k = MJ_EEPROM_CONTENT_12_START; s = MJ_EEPROM_CONTENT_12_SIZE; break;
          case 13: k = MJ_EEPROM_CONTENT_13_START; s = MJ_EEPROM_CONTENT_13_SIZE; break;
          case 14: k = MJ_EEPROM_CONTENT_14_START; s = MJ_EEPROM_CONTENT_14_SIZE; break;
        }
        Serial.print(F("String "));
        Serial.print(j + 1);
        Serial.print(F(" Size:"));
        Serial.print(mj_usedbytes[j]);
        Serial.print(F(":"));
        for (i = 0; i < mj_usedbytes[j] && i < s; i++) {
          Serial.write((char) mj_eeprom_read_byte(i + k));
        }
        Serial.println();
      }
      break;
    case 'k':
      putfifo_string(c + 1, 0);
      break;
    case 'w':
      if (num >= 2) {
        j = mj_convert_hex_nibble(c[1]);
        if (j >= 1 && j <= 14) {
          switch (j) {
            case 1: k = MJ_EEPROM_CONTENT_1_START; s = MJ_EEPROM_CONTENT_1_SIZE; break;
            case 2: k = MJ_EEPROM_CONTENT_2_START; s = MJ_EEPROM_CONTENT_2_SIZE; break;
            case 3: k = MJ_EEPROM_CONTENT_3_START; s = MJ_EEPROM_CONTENT_3_SIZE; break;
            case 4: k = MJ_EEPROM_CONTENT_4_START; s = MJ_EEPROM_CONTENT_4_SIZE; break;
            case 5: k = MJ_EEPROM_CONTENT_5_START; s = MJ_EEPROM_CONTENT_5_SIZE; break;
            case 6: k = MJ_EEPROM_CONTENT_6_START; s = MJ_EEPROM_CONTENT_6_SIZE; break;
            case 7: k = MJ_EEPROM_CONTENT_7_START; s = MJ_EEPROM_CONTENT_7_SIZE; break;
            case 8: k = MJ_EEPROM_CONTENT_8_START; s = MJ_EEPROM_CONTENT_8_SIZE; break;
            case 9: k = MJ_EEPROM_CONTENT_9_START; s = MJ_EEPROM_CONTENT_9_SIZE; break;
            case 10: k = MJ_EEPROM_CONTENT_10_START; s = MJ_EEPROM_CONTENT_10_SIZE; break;
            case 11: k = MJ_EEPROM_CONTENT_11_START; s = MJ_EEPROM_CONTENT_11_SIZE; break;
            case 12: k = MJ_EEPROM_CONTENT_12_START; s = MJ_EEPROM_CONTENT_12_SIZE; break;
            case 13: k = MJ_EEPROM_CONTENT_13_START; s = MJ_EEPROM_CONTENT_13_SIZE; break;
            case 14: k = MJ_EEPROM_CONTENT_14_START; s = MJ_EEPROM_CONTENT_14_SIZE; break;
          }
          for (i = 0; (c[i + 2] > 0) && ( i < (num - 1) ) && (i < s); i++) {
            mj_eeprom_write_byte(k + i, c[i + 2]);
          }
          mj_eeprom_write_byte(k + i, '\0');
          mj_usedbytes[j - 1] = i;
          mj_eeprom_update_value(k - 1, 1, i);
        }
      }
      break;
    case 't':
      if (num == 2) {
        j = mj_convert_hex_nibble(c[1]);
        if (j > 0) {
          putfifo_string(c, j);
        }
      }
      break;
    case 'd':
      if (num == 2) {
        switch (c[1]) {
          case 'r': // reset eeprom defaults;
            mj_keydelay[0] = MJ_DEFAULT_KEYDELAY_0;
            mj_keydelay[1] = MJ_DEFAULT_KEYDELAY_1;
            mj_keydelay[2] = MJ_DEFAULT_KEYDELAY_2;
            mj_keydelay[3] = MJ_DEFAULT_KEYDELAY_3;
            mj_keydelay[4] = MJ_DEFAULT_KEYDELAY_4;
            mj_keydelay[5] = MJ_DEFAULT_KEYDELAY_5;
            mj_keydelay[6] = MJ_DEFAULT_KEYDELAY_6;
            mj_keydelay[7] = MJ_DEFAULT_KEYDELAY_7;
            mj_keydelay[8] = MJ_DEFAULT_KEYDELAY_8;
            mj_keydelay[9] = MJ_DEFAULT_KEYDELAY_9;
            [[fallthrough]];
          case 's': // show delays
            Serial.print(F("KeyType Delay List"));
            for (i = 0; i < 10; i++) {
              Serial.write(' ');
              Serial.print(i);
              Serial.print(F(":"));
              Serial.print(mj_keydelay[i]);
            }
            Serial.println();
            break;
          case 'S':
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_0, mj_keydelay[0]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_1, mj_keydelay[1]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_2, mj_keydelay[2]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_3, mj_keydelay[3]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_4, mj_keydelay[4]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_5, mj_keydelay[5]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_6, mj_keydelay[6]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_7, mj_keydelay[7]);
            mj_eeprom_write_long(MJ_EEPROM_KEYDELAY_8, mj_keydelay[8]);
            mj_eeprom_update_value(MJ_EEPROM_KEYDELAY_9, 4, mj_keydelay[9]);
            break;
        }
      } else if (num >= 3) {
        if (c[1] >= '0' && c[1] <= '9') {
          i = c[1] - '0';
          value = mj_convert_dec_long(c + 2);
          mj_keydelay[i] = value;
        }
      }
    default:
      break;
  }
}

// Command atb
// atb<1-4><decimal> set button 4 to pin <decimal> - may not be higher than available pins
// atbr - reset button config to default
// atbs - show active button config
// atbS - save active button config to EEPROM
// atba... - Configure/show Button Actions
void parsebuttons(char *c, unsigned int num) {
  unsigned int value;
  int n = 0, i = 0;
  bool x;
  if (num >= 1) {
    switch (c[0]) {
      case '1':
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
        value = (unsigned int) mj_convert_dec_long(c + 1);
        i = c[0] - '1';
        // Check collision with other Pins already used (Key Inputs)
        x = true;
        for (n = 0; n < 4; n++) {
          if ( n != i ) {
            if (value == mj_keypin[n]) {
              x = false;
            }
          }
        }
        if (value == mj_ledpin1) {
          x = false;
        }
        if (value == mj_ledpin2) {
          x = false;
        }
#ifndef LED_BUILTIN_RX   // For example Arduino Uno - if we'd use the RX/TX Pins for Keys here we kill the serial communication. Leonardo works with 0/1
        if (value == 0 || value == 1) {
          x = false;
        }
#endif
        if (x == true && value < NUM_DIGITAL_PINS ) {
          pinMode(value, INPUT_PULLUP);
          mj_keypin[i] = value;
        } else if ( value == 0xff ) {
          mj_keypin[i] = 0xff;
        }
        break;
      case 'r': // Restore Key Default
        if ( (MJ_DEFAULT_KEYPIN_1 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_1 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_2 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_2 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_3 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_3 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_4 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_4 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_5 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_5 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_6 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_6 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_7 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_7 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_8 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_8 != mj_ledpin2 ) &&
             (MJ_DEFAULT_KEYPIN_9 != mj_ledpin1) && (MJ_DEFAULT_KEYPIN_9 != mj_ledpin2 ) ) {
          mj_keypin[0] = MJ_DEFAULT_KEYPIN_1;
          mj_keypin[1] = MJ_DEFAULT_KEYPIN_2;
          mj_keypin[2] = MJ_DEFAULT_KEYPIN_3;
          mj_keypin[3] = MJ_DEFAULT_KEYPIN_4;
          mj_keypin[4] = MJ_DEFAULT_KEYPIN_5;
          mj_keypin[5] = MJ_DEFAULT_KEYPIN_6;
          mj_keypin[6] = MJ_DEFAULT_KEYPIN_7;
          mj_keypin[7] = MJ_DEFAULT_KEYPIN_8;
          mj_keypin[8] = MJ_DEFAULT_KEYPIN_9;
        }
        for (n = 0; n < 10; n++) {
          if (mj_keypin[n] != 0xff ) {
            pinMode(mj_keypin[n], INPUT_PULLUP);
          }
        }
        [[fallthrough]];
      case 's': // Show serial Pin Assignment
        Serial.print(F("KEYPIN "));
        for (n = 0; n < 9; n++) {
          Serial.print(n + 1);
          Serial.write(':');
          if (mj_keypin[n] != 0xff) {
            Serial.print(mj_keypin[n]);
          } else {
            Serial.print(F("dis"));
          }
          Serial.write(' ');
        }
        Serial.println();
        break;
      case 'S': // Save Serial Pin Assignment to EEPROM
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_1, mj_keypin[0]); // Save Button 1
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_2, mj_keypin[1]); // Save Button 2
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_3, mj_keypin[2]); // Save Button 3
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_4, mj_keypin[0]); // Save Button 4
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_5, mj_keypin[1]); // Save Button 5
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_6, mj_keypin[2]); // Save Button 6
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_7, mj_keypin[0]); // Save Button 7
        mj_eeprom_write_byte(MJ_EEPROM_KEYPIN_8, mj_keypin[1]); // Save Button 8
        mj_eeprom_update_value(MJ_EEPROM_KEYPIN_9, 2, mj_keypin[8]); // Save Pin 9 plus recalculate + Update CRC
        break;
      case 'a': // Show/Configure Button Actions
        if (c[1] == 'x' && num >= 3 ) {
          value = (unsigned int) mj_convert_dec_long(c + 2);
          if (value > 0) {
            mj_fifo_a_putelement(value);
          }
        } else if (num > 4) {
          switch (c[1]) {
            case '1':
            case '2':
            case '3':
            case '4':
            case '5':
            case '6':
            case '7':
            case '8':
            case '9':
              value = (byte) mj_convert_dec_long(c + 4);
              if ((c[3] == '1' || c[3] == '2') && (c[2] == 'p' || c[2] == 'r')) {
                mj_action[c[1] - '1'][(c[2] == 'p') ? 0 : 1][c[3] - '1'] = value;
              }
              break;
          }
        } else if (num == 2) {
          switch (c[1]) {
            case 's': // Show Configured Actions
              for (i = 0; i < 9 ; i++) {
                Serial.print(F("Button "));
                Serial.print(i + 1);
                Serial.print(F(" press: "));
                if (mj_action[i][0][0] < 100) {
                  Serial.write(' ');
                }
                if (mj_action[i][0][0] < 10) {
                  Serial.write(' ');
                }
                Serial.print(mj_action[i][0][0]);
                Serial.print(F(" "));
                if (mj_action[i][0][1] < 100) {
                  Serial.write(' ');
                }
                if (mj_action[i][0][1] < 10) {
                  Serial.write(' ');
                }
                Serial.print(mj_action[i][0][1]);
                Serial.print(F(" release: "));
                if (mj_action[i][1][0] < 100) {
                  Serial.write(' ');
                }
                if (mj_action[i][1][0] < 10) {
                  Serial.write(' ');
                }
                Serial.print(mj_action[i][1][0]);
                Serial.print(F(" "));
                if (mj_action[i][1][1] < 100) {
                  Serial.write(' ');
                }
                if (mj_action[i][1][1] < 10) {
                  Serial.write(' ');
                }
                Serial.println(mj_action[i][1][1]);
              }
              break;
            case 'r':
              mj_action[0][0][0] = MJ_DEFAULT_KEY_1_PRESS_ACTION_1;
              mj_action[0][0][1] = MJ_DEFAULT_KEY_1_PRESS_ACTION_2;
              mj_action[0][1][0] = MJ_DEFAULT_KEY_1_RELEASE_ACTION_1;
              mj_action[0][1][1] = MJ_DEFAULT_KEY_1_RELEASE_ACTION_2;
              mj_action[1][0][0] = MJ_DEFAULT_KEY_2_PRESS_ACTION_1;
              mj_action[1][0][1] = MJ_DEFAULT_KEY_2_PRESS_ACTION_2;
              mj_action[1][1][0] = MJ_DEFAULT_KEY_2_RELEASE_ACTION_1;
              mj_action[1][1][1] = MJ_DEFAULT_KEY_2_RELEASE_ACTION_2;
              mj_action[2][0][0] = MJ_DEFAULT_KEY_3_PRESS_ACTION_1;
              mj_action[2][0][1] = MJ_DEFAULT_KEY_3_PRESS_ACTION_2;
              mj_action[2][1][0] = MJ_DEFAULT_KEY_3_RELEASE_ACTION_1;
              mj_action[2][1][1] = MJ_DEFAULT_KEY_3_RELEASE_ACTION_2;
              mj_action[3][0][0] = MJ_DEFAULT_KEY_4_PRESS_ACTION_1;
              mj_action[3][0][1] = MJ_DEFAULT_KEY_4_PRESS_ACTION_2;
              mj_action[3][1][0] = MJ_DEFAULT_KEY_4_RELEASE_ACTION_1;
              mj_action[3][1][1] = MJ_DEFAULT_KEY_4_RELEASE_ACTION_2;
              mj_action[4][0][0] = MJ_DEFAULT_KEY_5_PRESS_ACTION_1;
              mj_action[4][0][1] = MJ_DEFAULT_KEY_5_PRESS_ACTION_2;
              mj_action[4][1][0] = MJ_DEFAULT_KEY_5_RELEASE_ACTION_1;
              mj_action[4][1][1] = MJ_DEFAULT_KEY_5_RELEASE_ACTION_2;
              mj_action[5][0][0] = MJ_DEFAULT_KEY_6_PRESS_ACTION_1;
              mj_action[5][0][1] = MJ_DEFAULT_KEY_6_PRESS_ACTION_2;
              mj_action[5][1][0] = MJ_DEFAULT_KEY_6_RELEASE_ACTION_1;
              mj_action[5][1][1] = MJ_DEFAULT_KEY_6_RELEASE_ACTION_2;
              mj_action[6][0][0] = MJ_DEFAULT_KEY_7_PRESS_ACTION_1;
              mj_action[6][0][1] = MJ_DEFAULT_KEY_7_PRESS_ACTION_2;
              mj_action[6][1][0] = MJ_DEFAULT_KEY_7_RELEASE_ACTION_1;
              mj_action[6][1][1] = MJ_DEFAULT_KEY_7_RELEASE_ACTION_2;
              mj_action[7][0][0] = MJ_DEFAULT_KEY_8_PRESS_ACTION_1;
              mj_action[7][0][1] = MJ_DEFAULT_KEY_8_PRESS_ACTION_2;
              mj_action[7][1][0] = MJ_DEFAULT_KEY_8_RELEASE_ACTION_1;
              mj_action[7][1][1] = MJ_DEFAULT_KEY_8_RELEASE_ACTION_2;
              mj_action[8][0][0] = MJ_DEFAULT_KEY_9_PRESS_ACTION_1;
              mj_action[8][0][1] = MJ_DEFAULT_KEY_9_PRESS_ACTION_2;
              mj_action[8][1][0] = MJ_DEFAULT_KEY_9_RELEASE_ACTION_1;
              mj_action[8][1][1] = MJ_DEFAULT_KEY_9_RELEASE_ACTION_2;
              break;
            case 'S':
              for (n = 0; n < 9; n++) {
                for (i = 0; i < 2; i++) {
                  for (value = 0; value < 2; value++) {
                    mj_eeprom_write_byte(MJ_EEPROM_KEY_1_PRESS_ACTION_1 + ((n * 4) + (i * 2) + value), mj_action[n][i][value]);
                  }
                }
              }
              mj_eeprom_update_value(MJ_EEPROM_KEY_1_PRESS_ACTION_1, 1, mj_action[0][0][0]); // Re-calculate CRC...
              break;
          }
          break;
        }
        break;
    }
  }
}

// Jiggler options Parser
// atje enable Jiggler
// atjd disable Jiggler
// atjs show Jiggler state
// atjS Save Jiggler state
// atjr recover default Jiggler State
// atjt1<dezimal> Set Minimum Wait-Time between 2 Jiggles to number (0-65535) in ms
// atjt2<dezimal> Set Maximum Wait-Time between 2 Jiggles to number (0-65535) in ms
// atjts Show Jiggler timings
// atjtS Save Jiggler timings to EEPROM
// atjtr recover Jiggler Default State
// atjp<decimal> Set Pixel Max Pixel Movement to number (0-255)
// atjps Show Max Pixel Movement
// atjpS Save Max Pixel Movement to EEPROM
// atjpr recover Max Pixel Movement Default
void parsejiggler(char *c, unsigned int num) {
  unsigned int value;
  switch (c[0]) {
    case 'e':
      mj_jiggler_enable = 1;
      mj_ledtiming_changed = true;
      break;
    case 'd':
      mj_jiggler_enable = 0;
      mj_ledtiming_changed = true;
      break;
    case 'r':
      mj_jiggler_enable = MJ_DEFAULT_JIGGLER_ENABLE;
      mj_ledtiming_changed = true;
      [[fallthrough]];
    case 's':
      Serial.print(F("Jiggler "));
      Serial.print(mj_jiggler_enable ? "En" : "Dis");
      Serial.println(F("abled"));
      break;
    case 'S':
      mj_eeprom_update_value(MJ_EEPROM_JIGGLER_ENABLE, 1, mj_jiggler_enable & 0x01);
      break;
    case 't':
      if (num >= 2) {
        switch (c[1]) {
          case '1':
          case '2':
            if (num >= 3) {
              value = (unsigned int) mj_convert_dec_long(c + 2);
              if (c[1] == '1') {
                if (value <= mj_jiggler_max) {
                  mj_jiggler_min = value;
                }
              } else {
                if (value >= mj_jiggler_min) {
                  mj_jiggler_max = value;
                }
              }
            }
            break;
          case 'r': // Restore Default Timings
            mj_jiggler_min = MJ_DEFAULT_JIGGLER_MIN;
            mj_jiggler_max = MJ_DEFAULT_JIGGLER_MAX;
            [[fallthrough]];
          case 's':
            Serial.print(F("Jiggler delay min: "));
            Serial.print(mj_jiggler_min);
            Serial.print(F(" max: "));
            Serial.println(mj_jiggler_max);
            break;
          case 'S':
            mj_eeprom_write_int(MJ_EEPROM_JIGGLER_MIN, mj_jiggler_min);
            mj_eeprom_update_value(MJ_EEPROM_JIGGLER_MAX, 2, mj_jiggler_max);
            break;
          default:
            break;
        }
      }
      break;
    case 'p': // Jiggler Pixel maximum moves
      if (num >= 2) {
        switch (c[1]) {
          case '0':
          case '1':
          case '2':
          case '3':
          case '4':
          case '5':
          case '6':
          case '7':
          case '8':
          case '9':
            value = (byte) mj_convert_dec_long(c + 1);
            mj_jiggler_pixel = value;
            break;
          case 'S':
            mj_eeprom_update_value(MJ_EEPROM_JIGGLER_PIXEL, 1, mj_jiggler_pixel);
            break;
          case 'r':
            mj_jiggler_pixel = MJ_DEFAULT_JIGGLER_PIXEL;
            [[fallthrough]];
          case 's':
            Serial.print(F("Jiggler Pixel max: "));
            Serial.println(mj_jiggler_pixel);
            break;
          default:
            break;
        }
      }
      break;
  }
}

// Writes Baud-rate of serial line as on a atmega32u4 the Serial.begin(9600) does not really set baud.
void parsebaud() {
  const char p[] = "NOEMS";
  const char s[][4] = { "1", "1.5", "2" };
  Serial.print(F("Baudrate is "));
  Serial.print(Serial.baud());
  Serial.print(F(" at "));
  Serial.print(Serial.numbits());
  Serial.print(F("/"));
  Serial.write(p[Serial.paritytype()]);
  Serial.print(F("/"));
  Serial.print(s[Serial.stopbits()]);
  Serial.print(F(" RTS: "));
  Serial.println(Serial.rts());
}

void print_help(const char * textstart, int tabsize) {
  int column = 0, firstspace = 0, plen = 0, offset = 0, firstline = 1;
  char c = 0;
  char prefix[8];
  do {
    c = pgm_read_byte(textstart + offset);
    if (c == ' ' || c == '\0') {
      break;
    }
    prefix[offset] = c;
    offset++;
  }
  while (offset < 7);
  plen = offset;
  prefix[offset] = '\0';
  offset++;
  while ((c = pgm_read_byte(textstart + offset)) > 0) {
    if (c == '\t') {
      firstspace = 1;
      if (column != 0) {
        column = 0;
        Serial.println();
      }
      for (; column < tabsize; column++) {
        Serial.write(' ');
      }
    } else if (firstspace == 0 && c == ' ') {
      firstspace = 1;
      for (; column < (tabsize - 2); column++) {
        Serial.write(' ');
      }
      Serial.write('-');
      Serial.write(' ');
    } else if (c == '\n') {
      firstspace = 0;
      column = 0;
      Serial.println();
    } else {
      if (column == 0) {
        Serial.print(prefix);
        column += plen;
        if (firstline == 1) {
          Serial.print(F("..."));
          column += 3;
          firstline = 0;
          firstspace = 1;
          for (; column < (tabsize - 3); column++) {
            Serial.write(' ');
          }
          Serial.print(F("-- "));
        }
      }
      Serial.write(c);
      column++;
    }
    offset++;
  }
}

// Simply prints the info message to serial console
void printinfo() {
  Serial.println(F(INFOSTRING));
}


void print_at_help() {
  static const char helptext[] PROGMEM = "at Possible Commands - type at<command>? for more help\n"
                                         "e EEPROM\n"
                                         "l LED\n"
                                         "i Version Info\n"
                                         "j MouseJiggler\n"
                                         "t AutoType\n"
                                         "b Input Buttons\n"
                                         "r Baud Rate\n"
                                         "?? Print full Help\n";
  print_help(helptext, 9);
}

void print_att_help() {
  static const char helptext[] PROGMEM = "att AutoType\n"
                                         "s show\n"
                                         "k<string> type given string to keyboard\n"
                                         "w<1-e><string> set string number (HEX) 1-e (1-14)"
                                         "\tfollowing Escape-Codes exist for the string:"
                                         "\t\\e -> ESC, \\n or \\r -> RETURN, \\t -> TAB, \\\\ -> \\,"
                                         "\t\\d<0-9> Set typing delay to 0-9, \\x<HH> Hex code <HH>,"
                                         "\t\\X<HH> same as \\x but without auto-remapping"
                                         "\t\\s<HH> \"Special Key\" no press/release - hold the key \\sff releases all keys\n"
                                         "t<1-e> type string <1-e> to keyboard\n"
                                         "d... TYPING DELAYS\n"
                                         "d<0-9><number> sets delay <0-9> to <number> (0-4294967295) in ms\n"
                                         "ds show\n"
                                         "dS Save\n"
                                         "dr res. def.\n"
                                         "o... OS Mode\n"
                                         "o<0-1> set 0 (Win/Dos/LinuxKonsole) or 1 (Linux GTK+)\n"
                                         "os show\n"
                                         "oS Save\n"
                                         "or res. def.\n"
                                         "a... ACTIONS\n"
                                         "at<number> set Startup Delay to <number> (0-4294967295) ms"
                                         "\tWill automatically perform the defined actions after given Delay"
                                         "\tThe Delay counts from Power-On of the Stick not of the PC! Be careful about this!\n"
                                         "a<1-2><num> set Action <num> for AutoType Action <1-2> - will only be active after PowerCycle!\n"
                                         "ad disable\n"
                                         "as show\n"
                                         "aS Save\n"
                                         "ar res. def.\n";
  print_help(helptext, 20);
}

void print_atj_help() {
  static const char helptext[] PROGMEM = "atj Mouse Jiggler\n"
                                         "e enable\n"
                                         "d disable\n"
                                         "s show\n"
                                         "S Save\n"
                                         "r res. def.\n"
                                         "p... PIXEL MOVE\n"
                                         "p<num> set maximal move 0-255\n"
                                         "ps show\n"
                                         "pS Save\n"
                                         "pr res. def.\n"
                                         "t... TIMING\n"
                                         "t1<number> minimal wait time\n"
                                         "t2<number> maximal wait time\n"
                                         "ts show\n"
                                         "tS Save\n"
                                         "tr res. def.\n";
  print_help(helptext, 16);
}

void print_atl_help() {
  static const char helptext[] PROGMEM = "atl LED\n"
                                         "p... PIN Assignment\n"
                                         "p1<num> set LED1 pin to <num> - 255 disables LED\n"
                                         "p2<num> set LED2 pin to <num>\n"
                                         "ps show\n"
                                         "pS Save\n"
                                         "pr res. def."
                                         "\tyou cannot assign a pin that is already used by Buttons ot another LED\n"
                                         "t... TIMING\n"
                                         "t<1-4><0-65535) set timing modifier <1-4> to <0-65535>"
                                         "\t10000 is default, higher is slower"
                                         "\tnot all timings are used at every LED Mode\n"
                                         "ts show\n"
                                         "tS Save\n"
                                         "tr res. def.\n"
                                         "m... MODE\n"
                                         "m<0-7> set\n"
                                         "ms show\n"
                                         "mS Save\n"
                                         "mr res. def.\n"
                                         "o... OPTION\n"
                                         "o<0-ffff> set. The lowest 2 Bits invert LED1 or 2\n"
                                         "os show\n"
                                         "oS Save\n"
                                         "or res. def.\n";
  print_help(helptext, 21);
}

void print_ate_help() {
  static const char helptext[] PROGMEM = "ate EEPROM\n"
                                         "c Calc. CRC and show\n"
                                         "C Calc. CRC and update CRC\n"
                                         "d dump\n"
                                         "i init. def.\n"
                                         "I wipe and init. def.\n"
                                         "z only wipe"
                                         "\t The above 3 Commands need to be given 3 times to execute\n"
                                         "w... WRITE\n"
                                         "wb<xxx><yy> byte <yy> to Address <xxx> - Data and Address in HEX\n"
                                         "wi<xxx><yyyy> int\n"
                                         "wl<xxx><y-yyyyyyyy> long\n"
                                         "r... READ\n"
                                         "rb<xxx> byte from Address <xxx>\n"
                                         "ri<xxx> int\n"
                                         "rl<xxx> long\n";
  print_help(helptext, 25);
}

void print_atb_help() {
  static const char helptext[] PROGMEM = "atb Button Assignment\n"
                                         "<1-9><number> sets button <1-9> to pin <number> - 255 disables button"
                                         "\tyou cannot assign a pin that is already used by LED or another button\n"
                                         "s show\n"
                                         "S Save\n"
                                         "r res. def.\n"
                                         "a... ACTIONs\n"
                                         "a<1-9><p|r><1-2><num> set key <1-9> press/release <p|r> action <1-2> to <num> (0-255)"
                                         "\tfind a detailed list in tasks.h Line 40ff.\n"
                                         "as show active Key Action bindings\n"
                                         "aS Save\n"
                                         "ar res. def.\n"
                                         "ax<num> execute Action <num> immediately\n";
  print_help(helptext, 27);
}

void print_atr_help() {
  static const char helptext[] PROGMEM = "atr Baudrate\n"
                                         "\tshows the baud-Rate the PC is connected with\n";
  print_help(helptext, 10);
}

// Parse string given via serial and pass thru different decoders
void parsestring(char *c, unsigned int num) {
  if (num > 2) {
    if (c[0] == 'a' && c[1] == 't') {  // String needs to start with 'at'
      if (num == 4) {
        if (c[3] == '?') {
          switch (c[2]) {
            case 'j':
              printinfo();
              print_atj_help();
              break;
            case 'l':
              printinfo();
              print_atl_help();
              break;
            case 'i':
              printinfo();
              print_at_help();
              break;
            case 'e':
              printinfo();
              print_ate_help();
              break;
            case 't':
              printinfo();
              print_att_help();
              break;
            case 'b':  // input buttons
              printinfo();
              print_atb_help();
              break;
            case 'r':  // print actual baud rate
              printinfo();
              print_atr_help();
              break;
            case '?':  // print actual baud rate
              printinfo();
              print_at_help();
              Serial.println();
              print_atj_help();
              Serial.println();
              print_atl_help();
              Serial.println();
              print_ate_help();
              Serial.println();
              print_att_help();
              Serial.println();
              print_atb_help();
              Serial.println();
              print_atr_help();
              break;
            default:
              printinfo();
              print_at_help();
              break;
          }
        }
      }
      switch (c[2]) {  // The following character decides which submodule will decode further
        case 'j':  // JIGGLER related
          parsejiggler(c + 3, num - 3);
          break;
        case 'l':  // LED related
          parseled(c + 3, num - 3);
          break;
        case 'i':  // Info Message
          printinfo();
          break;
        case 'e':  // EEPROM related
          parseeeprom(c + 3, num - 3);
          break;
        case 't':  // autotype related
          parsekeyboard(c + 3, num - 3);
          break;
        case 'b':  // input buttons
          parsebuttons(c + 3, num - 3);
          break;
        case 'r':  // print actual baud rate
          parsebaud();
          break;
        default:
          if (num == 3) {
            print_at_help();
          }
          break;
      }
    }
  } else {
    print_at_help();
  }
}
